align 16
MergeSort  PROC uses esi edi ebx Arr:DWORD, count:DWORD
    ; MergeSort by Jim Giordano
    ; This routine sorts by merging streams of already sorted numbers.  It
    ;  differs from a classic merge sort in that the streams will vary in size
    ;  depending upon the input data, and a table of stream information is kept
    ;  to aid rapid merging.

    maxstream=4 ; 5  6  7  8; maximum number of streams being merged each level.
    maxlevel=15 ;13 12 11 10; worst case maximum number of levels needed to sort
                    ; maxstream=4 is faster on Intel, 6 on AMD but close enough.
    isize = 12  ; size of each mrg info block in bytes
    local rcount,level  ;,xheap
    local tempedi,esiptr,firststream
    local mrgstrt,mrgend,nexta,nextx,highestlevel
    local first[maxlevel]   ; address within mrg array of start of each level
    local next[maxlevel]    ; mrg array address of next free spot for each level
    local last[maxlevel]    ; last possible spot in each level
    local mrg[maxlevel*maxstream*isize/4]  ; info for mergestreams routine  

    ; where L=level (up to maxlevel), n=stream number (up to maxstream)
    ; mrg(L,n,0) = [ebx+0] = current value under test
    ; mrg(L,n,1) = [ebx+4] = address of the value under test
    ; mrg(L,n,2) = [ebx+8] = one past address of the last value in this stream
    
    mov eax,count
    cmp eax,2
    jb nosort
    mov rcount,eax  ; remaining values to test

;   note: heap is preallocated so all tests use the same spot in memory.
;      be sure to adjust positions of code for fastest routine when using alloc
;    shl eax, 2  ; compute bytes of scratch heap size
;    ;invoke SysAllocStringByteLen,0,eax ; otherwise one of these would be used
;    inv HeapAlloc,hHeap,0,eax
;    mov xheap,eax

    xor ecx,ecx             ; level number (index into array)
    mov highestlevel,ecx    ; preset highest level so far
    lea eax,mrg             ; start of stream locations
    .repeat
        mov first[ecx],eax  ; address of start of each level
        mov next[ecx],eax   ; next available spot 
        add eax,(maxstream-1)*isize ; number of possible streams this level
        mov last[ecx],eax   ; last possible spot for this level
        add eax,isize
        add ecx,4           ; levels are incremented by 4 (bytes/dword)
    .until ecx == 4*maxlevel
    
    mov esi,Arr
    mov esiptr,esi  ; where to start reading values for merge to scratch heap
    mov nexta,esi   ; where to store when merging back onto Arr
    mov edi,xheap
    mov nextx,edi   ; where to store when merging to the scratch xheap

    comment * The procedure - 
    
    GetNewData:
      level=0
      put info for Streams into mrg array (upto maxstream or until end of data)
      SetupStreamMerge:
        setup pointers to buffers
        merge to either xheap for even levels or back to Arr for odd levels
        level=level+1, update mrg(level) values for this new level
        if higher than higest level so far, set highest level
        if this new level is full, merge it (goto SetupStreamMerge)
        if there are values left in Arr (rcount>0), goto GetNewData
        we are now out of new data, merge remaining streams:
        if this is not the highest level, merge this one up (SetupStreamMerge)
        if there is more than one stream in this level, merge them
        if current data sits in xheap, move to Arr
    *

    align 4
    GetNewData:  ; back to here to get new streams from Arr until end of Arr
        mov level,0         ; new streams are always level 0
        mov ebx,next[0]     ; next available spot to store info for stream found
        mov esi,esiptr      ; where new data starts

        align 4
        GetStream:          ; get a stream of data from Arr        
            mov eax,[esi]   ; get first value of stream
            mov [ebx+4],esi ; save start of stream
            mov [ebx],eax   ; save value
            add esi,4       ; check next value
            sub rcount,1    ; count down available values
            jz DoneNewData  ; out of values, all done, ebx was not incremented
            cmp eax,[esi]   ; check next value to see direction of sequence
            jg DoReversed   ; possible reversed sequence
            align 4         ; this one should already have been aligned
        DoNormal:           ; this section marks out an increasing stream
            mov eax,[esi]   ; get new larger value
            add esi,4       ; address of next value to test
            sub rcount,1    ; count down available values
            jz DoneNewData  ; out of new values, go merge what we have
            cmp [esi],eax   ; test new value
            jge DoNormal    ; ok, go save it and test another

            cmp ebx,last[0] ; have we filled up a level?
            je DoneNewData  ; all done getting new data for now
            mov [ebx+8],esi ; save stop point  
            add ebx,isize   ; set up to save next info for next stream
            jmp GetStream   ; go back and get another stream to merge
            align 4
        DoReversed:         ; this section marks out a decreasing stream
            mov eax,[esi]   ; get new smaller value
            add esi,4
            sub rcount,1
            jz dorev        ; out of new values, go reverse stream
            cmp [esi],eax   ; test new value
            jl  DoReversed  ; ok, go save and test another.  note: if equal,
            ;    have to start over to maintain original order (keep stable)
            
          dorev:            ; now reverse the values found
            mov [ebx],eax   ; save first value of new stream
            mov ecx,esi
            sub ecx,4       ; address of last value in this stream
            mov edi,[ebx+4] ; current start of stream
            .repeat
                mov eax,[edi]   ; swap stream end for end
                mov edx,[ecx]
                mov [ecx],eax
                mov [edi],edx
                add edi,4
                sub ecx,4
            .until edi>=ecx
            cmp rcount,0    ; if we are out of new values, go do the merge
            je DoneNewData

            cmp ebx,last[0] ; have we filled up a level?
            je DoneNewData  ; yes, time to merge
            mov [ebx+8],esi ; no, save the stop point = one past last value  
            add ebx,isize   ; where info for next stream will be stored
            jmp GetStream   ; go get another stream to merge
            
            align 4
      DoneNewData:
        mov [ebx+8],esi     ; save ending point (one past end of last stream)
        add ebx,isize
        mov esiptr,esi      ; save next to test
        mov next[0],ebx     ; save next free cell for testing below

        ; Merge Streams
        align 4
      SetupStreamMerge:
        mov ecx,level       ; time to merge the streams found in this level
        mov edi,nextx       ; assume merging to heap
        test ecx,4          ; level increments by 4 for indexing into arrays
        cmovnz edi,nexta    ; nope, moving back to Arr
        mov tempedi,edi     ; save where we will store the data
        mov edx,first[ecx]  ; location of mrg info, first stream in this level
        mov esi,[edx+4]     ; location of first value in this stream
        mov firststream,esi ; save for wrapup

        mov eax,next[ecx]   ; one past last stream in this level
        sub eax,isize       ; start last stream in this level
        mov ebx,first[ecx]  ; start of first stream in this level
        .if eax>ebx         ; is there more than one stream to merge?
            mov mrgstrt,ebx ; yes, save begin and end of streams in this level
            mov mrgend,eax  ; yes, get last loc containing info for merge
            ;align 4   ; note: mergestreams is the most time critical code
            nop       ; extra to make inner loop the fastest, especially on AMD
            nop
            nop
          mergestreams:     ; merge streams found in mrg info array
            mov ebx,mrgstrt ; start of info for first stream in this level
            mov ecx,ebx     ; save address of current smallest value
            mov eax,[ebx]   ; first value in first string
            .repeat         ; note: left esi and edx free here for string sorts
                add ebx,isize   ; compare to value in next stream
                cmp [ebx],eax
                cmovl eax,[ebx] ; if smaller, save it
                cmovl ecx,ebx   ; and also it's address
            .until ebx==mrgend  ; ebx will equal mrgend when we exit
            mov [edi],eax       ; now have smallest of the group, save it
            mov eax,[ecx+4]     ; get address of smallest value
            add edi,4
            add eax,4           ; address of next value to test in this stream
            .if eax!=[ecx+8]    ; are we done with this stream?
                mov [ecx+4],eax ; no, save the new address
                mov eax,[eax]   ; get the value
                mov [ecx],eax   ; save new value
                jmp mergestreams    ; and go compare everyone again
            .endif
            ; we ran out of values in this stream, move last stream into this
            ;  spot if it's not the last one
            ; ebx is now equal to the offset into mrg array of last stream used
            .if ebx != ecx      ; was the last stream the smallest?
                mov eax,[ebx]   ; no, mov last group into this used up position
                mov [ecx],eax
                mov eax,[ebx+4]
                mov [ecx+4],eax
                mov eax,[ebx+8]
                mov [ecx+8],eax
            .endif
            sub ebx,isize       ; there is now one less stream to test
            mov mrgend,ebx      ; save the address of the last stream this level
            cmp ebx,mrgstrt     ; are we done?
            ja  mergestreams    ; no, go merge remaining streams
                       
        .endif
        ; only one stream left, copy it to output
        mov esi,[ebx+4] ; location of current value last remaining stream
        .repeat
            mov eax,[esi]   ; get value
            add esi,4       ; location of next value
            mov [edi],eax   ; store
            add edi,4       ; etc
        .until esi==[ebx+8] ; is it past the last one?

        mov eax,firststream ; address of the first input stream
        mov ecx,level       
        test ecx,4          ; are we copying to Arr or xheap?
        .if zero?
            mov nextx,edi   ; update next store location
            .if eax<nexta   ; save the earliest freed up location
                mov nexta,eax
            .endif
        .else
            mov nexta,edi
            .if eax<nextx
                mov nextx,eax
            .endif
        .endif
        
        mov eax,mrgstrt
        mov next[ecx],eax   ; reset this level to zero streams
        add ecx,4           ; update next higher level info for streams merged
        mov level,ecx
        .if ecx>highestlevel
            mov highestlevel,ecx    ; save highest level for final merge
        .endif
        mov edx,next[ecx]   ; location of next store
        mov eax,tempedi     ; start of new stream just created
        mov [edx+4],eax
        mov [edx+8],edi     ; next output address = one past end of this stream
        mov eax,[eax]       ; copy the first value of the stream
        mov [edx],eax
        add edx,isize
        mov next[ecx],edx   ; where next new stream will be stored
        cmp edx,last[ecx]   ; if this level is now full, go merge these steams
        ja SetupStreamMerge
        cmp rcount,0        ; if there are values left to do, go do it
        jne GetNewData
        ; out of data, merge all leftover streams from here up to highest level
        cmp ecx,highestlevel    ; if not the highest level, merge it up
        jne SetupStreamMerge
        sub edx,isize       ; if more than one stream in this level, go merge    
        cmp edx,first[ecx]
        ja  SetupStreamMerge
        test ecx,4              ; is final data in xheap?
        jz completelydone       ; no, done
        mov esi,xheap           ; yes, copy back to Arr
        mov edi,Arr
        mov ecx,count
        rep movsd
     
completelydone:
;   note: heap is preallocated to make sure all tests have an equal opportunity
;    ;invoke SysFreeString, xheap
;    inv HeapFree,hHeap,0,xheap

nosort:
    ret

MergeSort  ENDP
mrgsize=$-MergeSort
